const std = @import("std");

///
/// Returns a hash-backed table type of `Value`s indexed by `Key` and using `key_context` as the key
/// context.
///
pub fn Hashed(comptime Key: type, comptime Value: type,
    comptime key_context: KeyContext(Key)) type {

    const Allocator = std.mem.Allocator;

    return struct {
        allocator: Allocator,
        load_limit: f32,
        buckets: []Bucket,
        filled: usize,

        ///
        /// A slot in the hash table.
        ///
        const Bucket = struct {
            maybe_entry: ?struct {
                key: Key,
                value: Value,
            } = null,

            maybe_next_index: ?usize = null,
        };

        ///
        /// Hash table type.
        ///
        const Self = @This();

        ///
        /// Deinitializes `self`, preventing any further use.
        ///
        pub fn deinit(self: *Self) void {
            self.allocator.free(self.buckets);

            self.buckets = &.{};
        }

        ///
        /// Initializes a [Self] using `allocator` as the memory allocation strategy.
        ///
        /// Returns a new [Self] value or an [Allocator.Error] if initializing failed.
        ///
        pub fn init(allocator: Allocator) Allocator.Error!Self {
            return Self{
                .buckets = try allocator.alloc(Bucket, 4),
                .filled = 0,
                .allocator = allocator,
                .load_limit = 0.75,
            };
        }

        ///
        /// Searches for `key` to delete it, returning the deleted value or `null` if no matching
        /// key was found.
        ///
        pub fn remove(self: Self, key: Key) ?Value {
            var bucket = &(self.buckets[@mod(key_context.hash(key), self.buckets.len)]);

            if (bucket.maybe_entry) |*entry| if (key_context.equals(entry.key, key)) {
                defer entry.value = null;

                self.filled -= 1;

                return entry.value;
            };

            while (bucket.maybe_next_index) |index| {
                bucket = &(self.buckets[index]);

                if (bucket.maybe_entry) |*entry| if (key_context.equals(entry.key, key)) {
                    defer entry.value = null;

                    self.filled -= 1;

                    return entry.value;
                };
            }

            return null;
        }

        ///
        /// Attempts to insert the value at `key` to be `value` in `self`, returning an
        /// [InsertError] if it fails.
        ///
        pub fn insert(self: *Self, key: Key, value: Value) InsertError!void {
            if (self.loadFactor() >= self.load_limit) {
                const old_buckets = self.buckets;

                defer self.allocator.free(old_buckets);

                self.buckets = try self.allocator.alloc(Bucket, old_buckets.len * 2);

                for (old_buckets) |bucket, index| self.buckets[index] = bucket;
            }

            var hash = @mod(key_context.hash(key), self.buckets.len);

            while (true) {
                const bucket = &(self.buckets[hash]);

                if (key_context.equals((bucket.maybe_entry orelse {
                    bucket.maybe_entry = .{
                        .key = key,
                        .value = value
                    };

                    self.filled += 1;

                    break;
                }).key, key)) return error.KeyExists;

                hash = @mod(hash + 1, self.buckets.len);
            }
        }

        ///
        /// Returns the current load factor of `self`, which is derived from the number of capacity
        /// that has been filled.
        ///
        pub fn loadFactor(self: Self) f32 {
            return @intToFloat(f32, self.filled) / @intToFloat(f32, self.buckets.len);
        }

        ///
        /// Searches for a value indexed with `key` in `self`, returning it or `null` if no matching
        /// entry was found.
        ///
        pub fn lookup(self: Self, key: Key) ?*Value {
            var bucket = &(self.buckets[@mod(key_context.hash(key), self.buckets.len)]);

            if (bucket.maybe_entry) |*entry|
                if (key_context.equals(entry.key, key)) return &entry.value;

            while (bucket.maybe_next_index) |index| {
                bucket = &(self.buckets[index]);

                if (bucket.maybe_entry) |*entry|
                    if (key_context.equals(entry.key, key)) return &entry.value;
            }

            return null;
        }
    };
}

///
/// [InsertError.KeyExists] occurs when an insertion was attempted on a table with a matching key
/// already present.
///
pub const InsertError = std.mem.Allocator.Error || error {
    KeyExists,
};

///
/// Returns a context type for handling `Key` as a key in a table, associating hashing and equality
/// behaviors to it.
///
pub fn KeyContext(comptime Key: type) type {
    return struct {
        hash: fn (Key) usize,
        equals: fn (Key, Key) bool,
    };
}

///
/// Tests if the contents of `this_string` lexically equals the contents of `that_string`.
///
fn equalsString(this_string: []const u8, that_string: []const u8) bool {
    return std.mem.eql(u8, this_string, that_string);
}

///
/// Hashes `string` into a hash value of `usize`.
///
fn hashString(string: []const u8) usize {
    var hash = @as(usize, 5381);

    for (string) |byte| hash = ((hash << 5) + hash) + byte;

    return hash;
}

///
/// A [KeyContext] for handling `[]const u8` types.
///
pub const string_context = KeyContext([]const u8){
    .hash = hashString,
    .equals = equalsString,
};
